import { HttpService } from '@nestjs/axios';
import { NamedRoute } from '../../domain/entities/named-route';
import { IGeorouter } from '../../domain/interfaces/georouter.interface';
import { GeorouterSettings } from '../../domain/types/georouter-settings.type';
import { Path } from '../../domain/types/path.type';
import { Injectable } from '@nestjs/common';
import {
  catchError,
  defer,
  forkJoin,
  from,
  lastValueFrom,
  map,
  mergeAll,
  mergeMap,
  toArray,
} from 'rxjs';
import { AxiosError } from 'axios';

@Injectable()
export class GraphhopperGeorouter implements IGeorouter {
  _url: string;
  _urlArgs: Array<string>;
  _withTime: boolean;
  _withPoints: boolean;
  _withDistance: boolean;
  _paths: Array<Path>;
  _httpService: HttpService;

  constructor(url: string, httpService: HttpService) {
    this._url = url + '/route?';
    this._httpService = httpService;
  }

  async route(
    paths: Array<Path>,
    settings: GeorouterSettings,
  ): Promise<Array<NamedRoute>> {
    this._setDefaultUrlArgs();
    this._setWithTime(settings.withTime);
    this._setWithPoints(settings.withPoints);
    this._setWithDistance(settings.withDistance);
    this._paths = paths;
    const routes = await this._getRoutes();
    console.log(routes.length);
    return routes;
  }

  _setDefaultUrlArgs(): void {
    this._urlArgs = [
      'vehicle=car',
      'weighting=fastest',
      'points_encoded=false',
    ];
  }

  _setWithTime(withTime: boolean): void {
    this._withTime = withTime;
    if (withTime) {
      this._urlArgs.push('details=time');
    }
  }

  _setWithPoints(withPoints: boolean): void {
    this._withPoints = withPoints;
    if (withPoints) {
      this._urlArgs.push('calc_points=false');
    }
  }

  _setWithDistance(withDistance: boolean): void {
    this._withDistance = withDistance;
    if (withDistance) {
      this._urlArgs.push('instructions=true');
    } else {
      this._urlArgs.push('instructions=false');
    }
  }

  async _getRoutes(): Promise<Array<NamedRoute>> {
    const routes = Promise.all(
      this._paths.map(async (path) => {
        const url: string = [
          this._getUrl(),
          '&point=',
          path.points
            .map((point) => [point.lat, point.lon].join())
            .join('&point='),
        ].join('');
        const res = await lastValueFrom(
          this._httpService.get(url).pipe(
            map((res) => res.data.paths[0].distance),
            catchError((error: AxiosError) => {
              throw new Error(error.message);
            }),
          ),
        );
        return <NamedRoute>{
          key: path.key,
          route: res,
        };
      }),
    );
    return routes;
    // const date1 = new Date();
    // const urls = this._paths.map((path) =>
    //   defer(() =>
    //     this._httpService
    //       .get(
    //         [
    //           this._getUrl(),
    //           '&point=',
    //           path.points
    //             .map((point) => [point.lat, point.lon].join())
    //             .join('&point='),
    //         ].join(''),
    //       )
    //       .pipe(map((res) => res.data.paths[0].distance)),
    //   ),
    // );
    // const observables = from(urls);
    // const routes = observables.pipe(mergeAll(7), toArray());
    // routes.subscribe(() => {
    //   const date2 = new Date();
    //   console.log(date2.getTime() - date1.getTime());
    // });
    // return [];
  }

  _getUrl(): string {
    return [this._url, this._urlArgs.join('&')].join('');
  }
}
